#define _GNU_SOURCE
#include <stdio.h>
#include <wchar.h>
#include <stdlib.h>
#include <stdarg.h>
#include <signal.h>
#include <string.h>
#include <locale.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>

#define PROTVER		18

#define MAXCLIENTS	64
#define MSGLENGTH	128
#define MAXNAME		10

struct client {
	int		id;
	int		status;
	int		socket;
	char		name[MAXNAME];
} client[MAXCLIENTS];

volatile int got_ctrl_c = 0;
volatile int got_alrm = 0;

char *listfile;

struct solution {
	struct solution	*next;
	wchar_t		*word;
	int		length;
	int		player;
	char		path[16];
};

struct game {
	int		running;
	int		evaluating;
	int		players;
	wchar_t		box[16];
	int		nsol;
	int		nsol_left;
	struct solution	*solutions;
	int		timer;
} game;

struct suggest {
	wchar_t		word[32];
	char		votes[MAXCLIENTS];
} *suggest = 0;

int nsuggest = 0, nsuggestalloc = 0;
int needvotes;

int gametime = 180;

int ndict = 0;
int ndictalloc = 0;
wchar_t **dict = 0;

#define HASHSIZE 4096

struct dicthash {
	struct dicthash	*next;
	wchar_t		*word;
} *dicthash[HASHSIZE];

int ncfreq = 0;
int ncfreqalloc = 0;
struct charfreq {
	wchar_t		ch;
	int		freq;
} *charfreq = 0;
int totalfreq = 0;

//   0  1  2  3
//   4  5  6  7
//   8  9 10 11
//  12 13 14 15

int neighbours[16][9] = {
	{1, 4, 5, -1},
	{0, 2, 4, 5, 6, -1},
	{1, 3, 5, 6, 7, -1},
	{2, 6, 7, -1},

	{0, 1, 5, 8, 9, -1},
	{0, 1, 2, 4, 6, 8, 9, 10, -1},
	{1, 2, 3, 5, 7, 9, 10, 11, -1},
	{2, 3, 6, 10, 11, -1},
	
	{4, 5, 9, 12, 13, -1},
	{4, 5, 6, 8, 10, 12, 13, 14, -1},
	{5, 6, 7, 9, 11, 13, 14, 15, -1},
	{6, 7, 10, 14, 15, -1},

	{8, 9, 13, -1},
	{8, 9, 10, 12, 14, -1},
	{9, 10, 11, 13, 15, -1},
	{10, 11, 14, -1}
};

#define CS_NOTCONNECTED		0
#define CS_AUTHENTICATING0	1
#define CS_AUTHENTICATING1	2
#define CS_LOBBY		3
#define CS_PLAYING		4

void msg(struct client *c, char *msg, ...) {
	char buf[MSGLENGTH+1];
	va_list va;

	va_start(va, msg);
	memset(buf, 0, sizeof(buf));
	vsnprintf(buf, sizeof(buf), msg, va);
	va_end(va);
	printf("Sending '%s' to %s\n", buf, c->name);
	write(c->socket, buf, MSGLENGTH);
}

void broadcast(char *msg, ...) {
	char buf[MSGLENGTH+1];
	va_list va;
	int i;

	va_start(va, msg);
	memset(buf, 0, sizeof(buf));
	vsnprintf(buf, sizeof(buf), msg, va);
	va_end(va);
	printf("Broadcasting '%s' (%d bytes)\n", buf, strlen(buf));
	for(i = 0; i < MAXCLIENTS; i++) {
		if(client[i].status == CS_LOBBY || client[i].status == CS_PLAYING) {
			write(client[i].socket, buf, MSGLENGTH);
		}
	}
}

ssize_t forceread(int fd, char *ptr, size_t len) {
	ssize_t res;
	size_t left = len;

	do {
		res = read(fd, ptr, left);
		if(res == 0) return 0;
		if(res < 0) return res;
		left -= res;
		ptr += res;
	} while(left);
	return len;
}

void incfreq(wchar_t ch) {
	int i;
	struct charfreq *buf;

	totalfreq++;

	for(i = 0; i < ncfreq; i++) {
		if(charfreq[i].ch == ch) {
			charfreq[i].freq++;
			return;
		}
	}

	if(ncfreq >= ncfreqalloc) {
		buf = malloc(sizeof(struct charfreq) * ncfreqalloc * 2);
		if(!buf) {
			perror("malloc");
			exit(1);
		}
		memcpy(buf, charfreq, sizeof(struct charfreq) * ncfreqalloc);
		free(charfreq);
		charfreq = buf;
		ncfreqalloc *= 2;
	}

	charfreq[ncfreq].ch = ch;
	charfreq[ncfreq].freq = 1;
	ncfreq++;
}

void addword(wchar_t *word) {
	wchar_t **buf;
	int i;
	uint32_t h;
	struct dicthash *dh;

	h = 0;
	for(i = 0; word[i]; i++) h += word[i] * (i + 1);
	h %= HASHSIZE;

	for(dh = dicthash[h]; dh; dh = dh->next) {
		if(!wcscmp(word, dh->word)) return;
	}

	if(ndict >= ndictalloc) {
		buf = malloc(sizeof(wchar_t *) * ndictalloc * 2);
		if(!buf) {
			perror("malloc");
			exit(1);
		}
		memcpy(buf, dict, sizeof(wchar_t *) * ndictalloc);
		free(dict);
		dict = buf;
		ndictalloc *= 2;
	}
	dict[ndict] = wcsdup(word);

	dh = malloc(sizeof(struct dicthash));
	if(!dh) {
		perror("malloc");
		exit(1);
	}
	dh->next = dicthash[h];
	dh->word = dict[ndict];
	dicthash[h] = dh;

	ndict++;
	for(i = 0; i < wcslen(word); i++) {
		incfreq(word[i]);
	}
}

void sighandler(int sig) {
	switch(sig) {
		case SIGINT:
			got_ctrl_c = 1;
			break;
		case SIGALRM:
			if(game.timer) 
				game.timer--;
			else 
				setitimer(ITIMER_REAL, 0, 0);	
			got_alrm = 1;
			signal(SIGALRM, sighandler);
			break;
	}
}

int try_sol_sub(int wpos, int bpos, char *path, char *flags, wchar_t *w) {
	int i, j;

	path[wpos] = bpos;
	flags[bpos] = 1;
	if(!*w) return 1;
	for(i = 0; i < 16; i++) {
		j = neighbours[bpos][i];
		if(j < 0) break;
		if(!flags[j] && game.box[j] == *w) {
			if(try_sol_sub(wpos + 1, j, path, flags, w + 1)) {
				return 1;
			}
		}
	}
	flags[bpos] = 0;
	return 0;
}

int try_solution(wchar_t *w, int addflag) {
	char flags[16];
	char path[16];
	int i, res;
	struct solution *sol;

	memset(flags, 0, 16);
	for(i = 0; i < 16; i++) {
		if(game.box[i] == *w) {
			if(try_sol_sub(0, i, path, flags, w + 1)) {
				if(addflag) {
					sol = malloc(sizeof(struct solution));
					if(!sol) {
						perror("malloc");
						exit(1);
					}
					sol->word = w;
					sol->length = wcslen(w);
					sol->player = -1;
					memcpy(sol->path, path, 16);
					sol->next = game.solutions;
					game.solutions = sol;
					game.nsol++;
				}
				return 1;
			}
		}
	}
	return 0;
}

void startnewgame() {
	int i, r, j, k;
	char buf[1024];
	struct solution *sol;
	struct itimerval itmr;

	game.running = 1;
	game.nsol = 0;
	game.solutions = 0;
	game.players = 0;
	game.evaluating = 0;

	while(!game.nsol) {
		for(i = 0; i < 16; i++) {
			r = lrand48() % totalfreq;
			k = 0;
			for(j = 0; k + charfreq[j].freq <= r; j++) {
				k += charfreq[j].freq;
			}
			game.box[i] = charfreq[j].ch;
		}
		for(i = 0; i < ndict; i++) {
			try_solution(dict[i], 1);
		}
	}

	printf("Starting new game:\n");
	for(i = 0; i < 4; i++) {
		printf("   ");
		for(j = 0; j < 4; j++) {
			printf("%lc", game.box[i * 4 + j]);
			for(k = wcwidth(game.box[i * 4 + j]); k < 3; k++) printf(" ");
		}
		printf("\n");
	}

	broadcast("gametime %d", gametime);

	snprintf(buf, sizeof(buf), "game %d ", game.nsol);
	j = strlen(buf);
	setlocale(LC_CTYPE, "en_US.utf8");
	for(i = 0; i < 16; i++) {
		k = wctomb(buf + j, game.box[i]);
		if(k > 0) j += k;
	}
	buf[j] = 0;
	setlocale(LC_CTYPE, "");
	broadcast(buf);

	for(i = 0; i < MAXCLIENTS; i++) {
		if(client[i].status == CS_LOBBY) {
			client[i].status = CS_PLAYING;
		}
		if(client[i].status == CS_PLAYING) game.players++;
	}

	game.nsol_left = game.nsol;

	game.timer = gametime;
	itmr.it_interval.tv_sec = 1;
	itmr.it_interval.tv_usec = 0;
	itmr.it_value.tv_sec = 1;
	itmr.it_value.tv_usec = 0;
	setitimer(ITIMER_REAL, &itmr, 0);
}

void gameover(int timeoutflag) {
	struct solution *sol;
	char buf[1024];
	int i;

	game.evaluating = 1;

	if(timeoutflag) {
		printf("Time over\n");
		broadcast("timeout");
		setlocale(LC_CTYPE, "en_US.utf8");
		for(sol = game.solutions; sol; sol = sol->next) {
			if(sol->player == -1) {
				if(wcstombs(buf, sol->word, sizeof(buf)) > 0) {
					broadcast("missed %s", buf);
				}
			}
		}
		setlocale(LC_CTYPE, "");
	} else {
		printf("Puzzle solved");
		broadcast("solved");
	}
	needvotes = game.players;
	for(i = 0; i < nsuggest; i++) {
		setlocale(LC_CTYPE, "en_US.utf8");
		if(wcstombs(buf, suggest[i].word, sizeof(buf)) > 0) {
			broadcast("suggest %s", buf);
		}
		setlocale(LC_CTYPE, "");
	}
	broadcast("lastsuggest");
}

void handle_animreq(struct client *c, wchar_t *w) {
	struct solution *sol;
	char buf[17];
	int i;

	for(sol = game.solutions; sol; sol = sol->next) {
		if(!wcscmp(w, sol->word)) {
			for(i = 0; i < sol->length; i++) {
				buf[i] = sol->path[i] + 'a';
			}
			buf[sol->length] = 0;
			msg(c, "animate %s", buf);
		}
	}
}

void handle_vote(struct client *c, wchar_t *w) {
	int i;

	for(i = 0; i < nsuggest; i++) {
		if(!wcscmp(w, suggest[i].word)) {
			suggest[i].votes[c->id] = 1;
		}
	}
}

void handle_guess(struct client *c, wchar_t *w, char *mbstr) {
	struct solution *sol;
	int i, newsize;
	struct suggest *sugg;

	if(wcslen(w) > 16) return;

	for(sol = game.solutions; sol; sol = sol->next) {
		if(sol->player == -1 && !wcscmp(w, sol->word)) {
			sol->player = c->id;
			broadcast("contrib %s %s", c->name, mbstr);
			game.nsol_left--;
			if(!game.nsol_left) {
				gameover(0);
			}
			return;
		}
	}

	for(i = 0; i < nsuggest; i++) {
		if(!wcscmp(suggest[i].word, w)) {
			return;
		}
	}

	if(wcslen(w) > 2 && try_solution(w, 0)) {
		if(nsuggest >= nsuggestalloc) {
			newsize = nsuggestalloc * 2 + 1;
			sugg = malloc(sizeof(struct suggest) * newsize);
			if(!sugg) {
				perror("malloc");
				exit(1);
			}
			memcpy(sugg, suggest, sizeof(struct suggest) * nsuggest);
			free(suggest);
			suggest = sugg;
			nsuggestalloc = newsize;
		}
		wcscpy(suggest[nsuggest].word, w);
		memset(suggest[nsuggest].votes, 0, MAXCLIENTS);
		nsuggest++;
		printf("Added suggestion %ls\n", w);
	}
}

void addword_utf8(char *w) {
	size_t res;
	wchar_t buf[32];
	int i;

	res = mbstowcs(buf, w, 32);
	if(res != (size_t) -1) {
		if(wcslen(buf) > 2) addword(buf);
	}
}

void gamedone() {
	struct solution *sol, *nsol;
	int i, j, count;
	char buf[1024];
	FILE *f;

	game.running = 0;
	for(sol = game.solutions; sol; sol = nsol) {
		nsol = sol->next;
		free(sol);
	}
	for(i = 0; i < nsuggest; i++) {
		count = 0;
		for(j = 0; j < MAXCLIENTS; j++) {
			if(suggest[i].votes[j]) count++;
		}
		if(count >= needvotes && count) {
			printf("Adding new dictionary word '%ls'\n", suggest[i].word);
			f = fopen(listfile, "a");
			if(!f) {
				perror(listfile);
			} else {
				if(wcstombs(buf, suggest[i].word, sizeof(buf)) > 0) {
					fputs(buf, f);
					fputc(10, f);
					addword_utf8(buf);
				}
				fclose(f);
			}
		}
	}
	free(suggest);
	suggest = 0;
	nsuggest = 0;
	nsuggestalloc = 0;
}

void twit(struct client *c) {
	if(c->status == CS_PLAYING) game.players--;
	if(game.running && !game.players) {
		gamedone();
	}
	c->status = CS_NOTCONNECTED;
	broadcast("disconnect %s", c->name);
}

void parsemsg(struct client *c, char *m) {
	char *s1;
	int i, i1;
	wchar_t guess[17];

	printf("Got '%s' from %s\n", m, c->name);
	switch(c->status) {
		case CS_AUTHENTICATING0:
			if(1 == sscanf(m, "protocol %d", &i1)) {
				if(i1 == PROTVER) {
					msg(c, "proto ok");
					c->status = CS_AUTHENTICATING1;
				} else {
					msg(c, "proto bad");
					twit(c);
				}
			}
			break;
		case CS_AUTHENTICATING1:
			if(!strncmp(m, "nick ", 5) && (s1 = strdup(m + 5))) {
				if(strlen(s1) > MAXNAME - 1) {
					msg(c, "nick taken");
					return;
				}
				strncpy(c->name, s1, MAXNAME);
				c->name[MAXNAME - 1] = 0;
				for(i = 0; i < MAXNAME && c->name[i]; i++) {
					if(c->name[i] <= ' ' || c->name[i] >= 127) {
						msg(c, "nick taken");
						return;
					}
				}
				free(s1);
				for(i = 0; i < MAXCLIENTS; i++) {
					if(&client[i] != c
					&& client[i].status != CS_NOTCONNECTED
					&& !strcmp(c->name, client[i].name)) {
						break;
					}
				}
				if(i == MAXCLIENTS) {
					msg(c, "nick ok");
					if(game.running && game.evaluating) {
						msg(c, "evaluating");
					}
					for(i = 0; i < MAXCLIENTS; i++) {
						if(client[i].status == CS_LOBBY) {
							msg(c, "in lobby %s", client[i].name);
						} else if(client[i].status == CS_PLAYING) {
							msg(c, "playing %s", client[i].name);
						}
					}
					c->status = CS_LOBBY;
					broadcast("in lobby %s", c->name);
				} else {
					msg(c, "nick taken");
				}
			}
			break;
		case CS_LOBBY:
			if(!strcmp(m, "request new")) {
				if(!game.running) {
					startnewgame();
				}
			}
			break;
		case CS_PLAYING:
			if(!game.evaluating && (1 == sscanf(m, "guess %a[^\n]", &s1))) {
				guess[16] = 0;
				setlocale(LC_CTYPE, "en_US.utf8");
				if(mbstowcs(guess, s1, 16) > 0) {
					handle_guess(c, guess, s1);
				}
				setlocale(LC_CTYPE, "");
				free(s1);
			} else if(game.evaluating && !strcmp(m, "in lobby")) {
				c->status = CS_LOBBY;
				broadcast("in lobby %s", c->name);
				game.players--;
				if(game.running && !game.players) {
					gamedone();
				}
			} else if(1 == sscanf(m, "animrequest %a[^\n]", &s1)) {
				guess[16] = 0;
				setlocale(LC_CTYPE, "en_US.utf8");
				if(mbstowcs(guess, s1, 16) > 0) {
					handle_animreq(c, guess);
				}
				setlocale(LC_CTYPE, "");
				free(s1);
			} else if(1 == sscanf(m, "vote %a[^\n]", &s1)) {
				guess[16] = 0;
				setlocale(LC_CTYPE, "en_US.utf8");
				if(mbstowcs(guess, s1, 16) > 0) {
					handle_vote(c, guess);
				}
				setlocale(LC_CTYPE, "");
				free(s1);
			}
			break;
	}
}

void loaddict() {
	FILE *f;
	char buf[1024];

	f = fopen(listfile, "r");
	if(!f) {
		perror(listfile);
		exit(1);
	}

	while(!feof(f)) {
		if(fgets(buf, sizeof(buf), f)) {
			if(strlen(buf) && buf[strlen(buf) - 1] == 10) buf[strlen(buf) - 1] = 0;
			if(strlen(buf)) addword_utf8(buf);
		}
	}

	fclose(f);
}

int main(int argc, char **argv) {
	struct sockaddr_in sockaddr;
	int sock, max, s;
	int running = 1;
	char *serverid = "unnamed";
	ssize_t size;
	char buf[MSGLENGTH + 1];
	fd_set fds;
	int i;
	socklen_t socklen;
	uint32_t addr;

	if(argc == 4) {
		serverid = argv[3];
		gametime = atoi(argv[2]);
	} else if(argc == 3) {
		gametime = atoi(argv[2]);
	} else if(argc != 2) {
		fprintf(stderr, "Usage: %s listfile [timeout [server_info]]\n", argv[0]);
		exit(1);
	}

	dict = malloc(sizeof(wchar_t *));
	ndictalloc = 1;
	charfreq = malloc(sizeof(struct charfreq));
	ncfreqalloc = 1;

	setlocale(LC_ALL, "");
	listfile = argv[1];
	loaddict();

	printf("%d words in dictionary\n", ndict);
	for(i = 0; i < ncfreq; i++) {
		if(i) printf(", ");
		printf("%lc: %d", charfreq[i].ch, charfreq[i].freq);
	}
	printf("\n");

	srand48(time(0));

	sockaddr.sin_family = AF_INET;
	sockaddr.sin_port = htons(1789);
	sockaddr.sin_addr.s_addr = INADDR_ANY;

	sock = socket(PF_INET, SOCK_STREAM, 0);
	if(sock < 0) {
		perror("socket");
		exit(1);
	}

	if(bind(sock, (struct sockaddr *) &sockaddr, sizeof(sockaddr))) {
		perror("bind");
		exit(1);
	}

	if(listen(sock, 5)) {
		perror("listen");
		exit(1);
	}

	printf("Server running at port 1789\n");

	signal(SIGINT, sighandler);
	signal(SIGALRM, sighandler);
	signal(SIGPIPE, SIG_IGN);

	for(i = 0; i < MAXCLIENTS; i++) {
		client[i].id = i;
	}

	while(running) {
		FD_ZERO(&fds);
		FD_SET(sock, &fds);
		max = sock;
		for(i = 0; i < MAXCLIENTS; i++) {
			if(client[i].status != CS_NOTCONNECTED) {
				FD_SET(client[i].socket, &fds);
				if(client[i].socket > max) max = client[i].socket;
			}
		}
		if(select(max + 1, &fds, 0, 0, 0) >= 0) {
			if(FD_ISSET(sock, &fds)) {
				socklen = sizeof(sockaddr);
				s = accept(sock, (struct sockaddr *) &sockaddr, &socklen);
				if(s < 0) {
					perror("accept");
				} else {
					for(i = 0; i < MAXCLIENTS; i++) {
						if(client[i].status == CS_NOTCONNECTED) break;
					}
					if(i == MAXCLIENTS) {
						close(s);
					} else {
						addr = ntohl(sockaddr.sin_addr.s_addr);
						printf("Connect from %d.%d.%d.%d\n",
							(addr >> 24) & 255,
							(addr >> 16) & 255,
							(addr >>  8) & 255,
							(addr >>  0) & 255);
						client[i].status = CS_AUTHENTICATING0;
						client[i].socket = s;
						strncpy(client[i].name, "<unknown>", MAXNAME);
						msg(&client[i], "welcome %s", serverid);
					}
				}
			}
			for(i = 0; i < MAXCLIENTS; i++) {
				if(client[i].status != CS_NOTCONNECTED
				&& FD_ISSET(client[i].socket, &fds)) {
					size = forceread(client[i].socket, buf, MSGLENGTH);
					buf[MSGLENGTH] = 0;
					if(size == 0) {
						printf("Disconnect from %s\n", client[i].name);
						close(client[i].socket);
						twit(&client[i]);
					} else if(size != MSGLENGTH) {
						msg(&client[i], "protocol error");
						printf("Protocol error from %s\n", client[i].name);
						close(client[i].socket);
						twit(&client[i]);
					} else {
						parsemsg(&client[i], buf);
					}
				}
			}
		}
		if(got_ctrl_c) running = 0;
		if(got_alrm) {
			got_alrm = 0;
			broadcast("timer %d", game.timer);
		}
		if(game.running && !game.evaluating && !game.timer) {
			gameover(1);
		}
	}

	printf("Terminating\n");

	for(i = 0; i < MAXCLIENTS; i++) {
		if(client[i].status != CS_NOTCONNECTED) {
			msg(&client[i], "server terminating");
			close(client[i].socket);
		}
	}

	return 0;
}
